///////////////////////////////////////////////////////////////////////////////
//             includes, defines, declarations etc.
///////////////////////////////////////////////////////////////////////////////
#include <string> // used by myipaddr.hh
#include <sstream> // used in this file
#include <iostream> // used in this file

// These are needed by mysocket.hh:
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>


#include "myipaddr.hh"
#include "mysocket.hh"
#include <jack/jack.h>
#include <jack/driver.h>
// This is needed for "jack_driver_become_client()"
#include <jack/internal.h>
#include "chain.hh"
#include "master_callbacks.hh"
#include "slave_driver.hh"

#include <sched.h> // needed for sched_yield().


///////////////////////////////////////////////////////////////////////////////
//                This is the actual class definition
///////////////////////////////////////////////////////////////////////////////

chain::chain (const char *chainName, const myIPAddress& remoteIP) \
							: name (chainName), \
						transmitterName (chainName), \
						receiverName (chainName) \
{

	thePeerIP = remoteIP;
	numberOfInputPorts = numberOfOutputPorts = 0;
	inputPorts = outputPorts = 0;
	inputSockets = 0;
	outputSockets = 0;
	transmitter = receiver = 0;
	blockingMode = false;

	// Define the name for the jack clients we will create
	transmitterName += ".trans";
	receiverName += ".rec";

	// Setup the clients and register with JACK
	createClients ();
	registerCallbacks();

	// Create the "dummy" ports. These are only needed on the
	// "master", but it doesn't hurt to have them on both.
	dummyTransmitterInPort = jack_port_register (transmitter, \
		"dumyin", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	dummyTransmitterOutPort = jack_port_register (transmitter, \
		"dumyout", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
	dummyReceiverInPort = jack_port_register (receiver, \
		"dumyin", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
	dummyReceiverOutPort = jack_port_register (receiver, \
		"dumyout", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

	// Tie the ports together; this way we don't have to worry about these
	// ports during our "process" callback. These ports only exist as a
	// hack to force jackd to order the processing graph the way we want.
	// FIXME
	// jack has a bug that prevents this from working; probably will be
	// fixed on the next version, since it's now acknowledged.
	//jack_port_tie (dummyTransmitterInPort, dummyTransmitterOutPort);
	//jack_port_tie (dummyReceiverInPort, dummyReceiverOutPort);
}

void
chain::createClients ()
{
	// FIXME - on error, memory leak; this should raise an exception
	if ((transmitter = jack_client_new (transmitterName.c_str())) == 0) {
		cerr << "jack server not running?" << endl;
		exit (1);
	}
	if ((receiver = jack_client_new (receiverName.c_str())) == 0) {
		cerr << "jack server not running?" << endl;
		exit (1);
	}
}

void
chain::registerCallbacks()
{
	///////////////////////////////////////////////////////////////////////
	// Tell JACK about the callbacks for the two clients:
	///////////////////////////////////////////////////////////////////////
	
	// The main "process" callback
	jack_set_process_callback (transmitter, ::transmit, \
						static_cast<void*> (this));
	jack_set_process_callback (receiver, ::receive, static_cast<void*> (this));

	// The "please note that the buffersize has changed" callback; Only one
	// of the clients needs to have the callback called, since we can (and
	// will) reconfigure the object with the callback.
	jack_set_buffer_size_callback (transmitter, ::changeBufsize, \
						static_cast<void*> (this));
	// jack_set_buffer_size_callback (receiver, 0, 0);

	// The "please note that the sample rate has changed" callback; Only one
	// of the clients needs to have the callback called, since we can (and
	// will) reconfigure the object with the callback.
	jack_set_sample_rate_callback (transmitter, ::changeRate, \
						static_cast<void*> (this));
	// jack_set_sample_rate_callback (receiver, 0, 0);

	// The "you're done" callback (jackd has exited or we've been removed
	// from the graph). We need to register both clients here, since JACK
	// may decide to eliminate only one of them from the graph; if this
	// happens, the other should be removed from the graph too, so each
	// callback destroys the client that *didn't* had the callback called.
	jack_on_shutdown (transmitter, ::destroyReceiver, \
						static_cast<void*> (this));
	jack_on_shutdown (receiver, ::destroyTransmitter, \
						static_cast<void*> (this));
}

void
chain::createPorts (const int numberOfInputPorts, \
			const int numberOfOutputPorts)
{

	this->numberOfInputPorts = numberOfInputPorts;
	this->numberOfOutputPorts = numberOfOutputPorts;

	inputPorts = new jack_port_t*[numberOfInputPorts];
	outputPorts = new jack_port_t*[numberOfOutputPorts];

	// Register input and output ports with JACK.
	// we need to name each port
	stringstream portname;
	for (int i = 0; i < numberOfInputPorts; i++) {
		portname << "in_" << i+1 << flush;
		inputPorts[i] = jack_port_register (transmitter, \
			portname.str().c_str(), JACK_DEFAULT_AUDIO_TYPE, \
			JackPortIsInput, 0);
		portname.str(""); // "clear" the string
	}
	for (int i = 0; i < numberOfOutputPorts; i++) {
		portname << "out_" << i+1 << flush;
		outputPorts[i] = jack_port_register (receiver, \
			portname.str().c_str(), JACK_DEFAULT_AUDIO_TYPE, \
			JackPortIsOutput, 0);
		portname.str(""); // "clear" the string
	}
}

void
chain::createOutputSockets (myIPAddress* sendToPorts)
{
	outputSockets = new udpWriteSocket<jack_default_audio_sample_t>* \
						[numberOfInputPorts];

	for (int i = 0; i < numberOfInputPorts; i++) {
		// If sendToPorts has only the ports set and not the
		// IP, no problem: we already know the peer IP, we
		// just really need to know the ports.
		sendToPorts[i].IPAddress (thePeerIP.IPAddress());
		outputSockets[i] = new \
			udpWriteSocket<jack_default_audio_sample_t> \
		(sendToPorts[i]);
	}
}

void
chain::createInputSockets (const myIPAddress* bindToPorts = 0)
{
	inputSockets = new udpReadSocket<jack_default_audio_sample_t>* \
						[numberOfOutputPorts];
	if (bindToPorts) {
		// If we were told were to bind to, obey
		for (int i = 0; i < numberOfOutputPorts; i++) {
			inputSockets[i] = new udpReadSocket\
	                        <jack_default_audio_sample_t> \
				(bindToPorts[i], blockingMode);
		}
	} else {
		// Nobody told us nothing, let's bind anywhere;
		// this is actually equivalent to "INADDR_ANY":
		myIPAddress IP;
		for (int i = 0; i < numberOfOutputPorts; i++) {
			inputSockets[i] = new udpReadSocket\
			<jack_default_audio_sample_t> (IP, blockingMode);
		}
	}
}

// If the user of the class needs to know to which ports we are binded,
// perhaps to tell the peer where to send packets, this is how to find out.
void
chain::listInputSockets (myIPAddress* bindedPorts) {
	for (int i = 0; i < numberOfOutputPorts; i++) {
		bindedPorts[i] = inputSockets[i]->IPAddress();
	}
}

void
chain::activate()
{
	// tell the JACK server that we are ready to roll
	// FIXME - on error, memory leak; this should raise exception
	if (jack_activate (transmitter)) {
		cerr << "cannot activate client" << endl;
		exit (1);
	}
	if (jack_activate (receiver)) {
		cerr << "cannot activate client" << endl;
		exit (1);
	}
}

int
chain::transmit (jack_nframes_t nframes)
{
	// Only try to transmit if we already set up the network sockets
	if (outputSockets) {
		jack_default_audio_sample_t* inputData;
		for (int i = 0; i < numberOfInputPorts; i++) {
			inputData = static_cast<jack_default_audio_sample_t *> \
				(jack_port_get_buffer (inputPorts[i], nframes));
			outputSockets[i]->write (inputData, nframes);
		}
		// if the other machine is in fact this one (which is the same
		// as saying, "if you really want to shoot yourself on the
		// foot").
		sched_yield();
	}
	return 0;
}

int
chain::receive (jack_nframes_t nframes)
{
	// Only try to receive if we already set up the network sockets
	if (inputSockets) {
		jack_default_audio_sample_t* outputData;
		for (int i = 0; i < numberOfOutputPorts; i++) {
			outputData = static_cast<jack_default_audio_sample_t *> \
				(jack_port_get_buffer (outputPorts[i], nframes));
			inputSockets[i]->read (outputData, nframes);
		}
		// if the other machine is in fact this one (which is the same
		// as saying, "if you really want to shoot yourself on the
		// foot").
		sched_yield();
	} else {
		// No sockets, so let's just output silence
		for (int i = 0; i < numberOfOutputPorts; i++) {
			memset(jack_port_get_buffer (outputPorts[i], nframes), \
							0, nframes * sizeof \
						(jack_default_audio_sample_t));
		}
	}
	return 0;
}


void
chain::disconnect ()
{
	jack_deactivate (receiver);
	jack_deactivate (transmitter);
}

chain::~chain()
{
	destroyReceiver();
	destroyTransmitter();
	unregisterJackPorts();

	// The sockets need to be deleted explicitly by us
	if (outputSockets != 0) {
		for (int i = 0; i < numberOfInputPorts; i++) {
			delete outputSockets[i];
		}
		delete[] outputSockets;
	}

	if (inputSockets != 0) {
		for (int i = 0; i < numberOfOutputPorts; i++) {
			delete inputSockets[i];
		}
		delete[] inputSockets;
	}
}

void
chain::destroyTransmitter ()
{
	// Aparently, jack_client_close should remove all of the client's
	// ports, but it doesn't; let's do it by hand here.
	unregisterJackPorts();
	jack_client_close (transmitter);
}

void
chain::destroyReceiver ()
{
	// Aparently, jack_client_close should remove all of the client's
	// ports, but it doesn't; let's do it by hand here.
	unregisterJackPorts();
	jack_client_close (receiver);
}

int
chain::changeRate (jack_nframes_t nframes) {
	return 0;
}

int
chain::changeBufsize (jack_nframes_t nframes) {
	return 0;
}

void
chain::unregisterJackPorts () {
	if (inputPorts) {
		for (int i = 0; i < numberOfInputPorts; i++) {
			jack_port_unregister (transmitter, inputPorts[i]);
		}
		delete[] inputPorts;
		inputPorts = 0;
	}
	if (outputPorts) {
		for (int i = 0; i < numberOfOutputPorts; i++) {
			jack_port_unregister (transmitter, outputPorts[i]);
		}
		delete[] outputPorts;
		outputPorts = 0;
	}
	jack_port_unregister (transmitter, dummyTransmitterInPort);
	jack_port_unregister (transmitter, dummyTransmitterOutPort);
	jack_port_unregister (receiver, dummyReceiverInPort);
	jack_port_unregister (receiver, dummyReceiverOutPort);
}

///////////////////////////////////////////////////////////////////////////////
//    The "slave" chain; supposedly it will be used on the slave driver
///////////////////////////////////////////////////////////////////////////////

slaveChain::slaveChain (const char *chainName, const myIPAddress& \
						remoteIP) \
			: chain::chain (chainName, remoteIP)
{
	blockingMode = true;
}

void
slaveChain::createClients ()
{
	// FIXME - on error, memory leak; this should raise an exception
#ifdef notdef
	// This code is disabled while I figure out how
	// to make it link correctly
	if ((transmitter = jack_driver_become_client (transmitterName.c_str())) == 0) {
		jack_error ("Unknown error trying to become a client");
		return;
	}
	if ((receiver = jack_driver_become_client (receiverName.c_str())) == 0) {
		jack_error ("Unknown error trying to become a client");
		return;
	}
#else
	return;
#endif
}

void
slaveChain::registerCallbacks()
{
	///////////////////////////////////////////////////////////////////////
	// This method does nothing; for the slave, the callbacks are
	// set up differently.
	///////////////////////////////////////////////////////////////////////
	return;
}
