#ifndef CHAIN_HH
#define CHAIN_HH

// This is the class that defines a complete "chain" of data, where
// one jack client receives data and sends it for remote processing
// and another receives the data back and delivers it back to jack.
// Both clients are part of this class.

// files that include this one must also include:
// myipaddr.hh, mysocket.hh, master_callbacks.hh, slave_driver.hh,
// <string>, <jack/jack.h> and the files needed by these.

class chain {
public:
	chain (const char* chainName, const myIPAddress& remoteIP);
	virtual ~chain();

	void createPorts (const int numberOfInputPorts, \
			const int numberOfOutputPorts);
	// FIXME This parameter should really be const, but can't be.
	void createOutputSockets (myIPAddress* sendToPorts);
	void createInputSockets (const myIPAddress* bindToPorts = 0);
	// fills the given array with the IP addresses/ports for the sockets;
	// use it if you need to know where th input sockets are binded to.
	void listInputSockets (myIPAddress* bindedPorts);

	// Activate the clients within JACK
	void activate();
	// Disconnect from the graph without destroying the clients or the ports
	void disconnect ();

	void destroyTransmitter ();
	void destroyReceiver ();

	inline myIPAddress peerIP() const;

	///////////////////////////////////////////////////////////////////////
	// These methods are called when JACK issues a callback on our clients.
	///////////////////////////////////////////////////////////////////////
	int transmit (jack_nframes_t nframes);
	int receive (jack_nframes_t nframes);
	int changeRate (jack_nframes_t nframes);
	int changeBufsize (jack_nframes_t nframes);

protected:
	// FIXME - most of these should be private
	bool blockingMode;
	string name, transmitterName, receiverName;
	jack_client_t *transmitter, *receiver;

	// To be used by the constructor
	virtual void createClients();
	virtual void registerCallbacks ();

	// The arrays of JACK ports where data arrives / is delivered to; the
	// input ports are used by the transmitter jack client and the output
	// ports are used by the receiver jack client. STL containers are
	// probably inadequate here, since we need access to these to be
	// as fast as possible.
	jack_port_t **inputPorts;
	jack_port_t **outputPorts;

	// The arrays for the sockets associated with the JACK in/out ports
	udpWriteSocket<jack_default_audio_sample_t>** outputSockets;
	udpReadSocket<jack_default_audio_sample_t>** inputSockets;

	// This stores the sizes of the arrays defined above.
	int numberOfOutputPorts, numberOfInputPorts;

	// These are the "dummy" ports that we'll use to force JACK to order
	// the processing graph in a way that is "good" for us, that is, the
	// transmitter client is on or near the "start" of the graph and the
	// receiver client is on or near the "end" of the graph. This is a hack.
	jack_port_t *dummyTransmitterInPort, *dummyTransmitterOutPort, \
		*dummyReceiverInPort, *dummyReceiverOutPort;
private:
	myIPAddress thePeerIP;
	void unregisterJackPorts ();
};

// FIXME - This inheritance relationship is totally wrong.
class slaveChain: public chain {
public:
	slaveChain (const char* chainName, const myIPAddress& \
					remoteIP);
private:
	// To be used by the constructor
	virtual void createClients();
	virtual void registerCallbacks ();
};

// Inline methods
inline myIPAddress 
chain::peerIP() const
{
	return thePeerIP;
}

#endif
