///////////////////////////////////////////////////////////////////////////////
// This is a simple program that reads network data in realtime and echoes
// it back. This is basically a test tool.
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
//           includes, defines, declarations etc.
///////////////////////////////////////////////////////////////////////////////

// "standard" stuff
#include <iostream>
#include <unistd.h>
#include <sys/types.h>

// networking and other stuff, mostly needed by mysocket.hh
#include <sys/socket.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <string>
#include "myipaddr.hh"
#include "mysocket.hh"

#include <getopt.h>
#include <sched.h> // realtime scheduling
#include <sys/mman.h> // mlockall()

#include <jack/types.h>

// Maximum network buffer size; TCP/IP limits this to 65535 bytes.
// In "The practise of programming", Kernighan & Pike suggest that
// "enum" is better than "#define".
enum { MAX_NET_BUFSIZE = 65535 };

// Since we don't process the program arguments inside main(), we need
// these to exchange data between parseargs() and main(); ugh! Passing
// pointers or references around doesn't seem better tho...
myIPAddress peerIP, *bindToPorts, *sendToPorts;
int channels;


///////////////////////////////////////////////////////////////////////////////
//                             Functions
///////////////////////////////////////////////////////////////////////////////

void
usage (char* progname) {
	cerr << "syntax: " << progname;
	cerr << " -channels number_of_channels ";
	cerr << "-remoteip ipnumber remote_ports local_ports"  << endl;
	cerr << "Note that the order of the ports on the command line" << endl;
	cerr << "should be the same as that of masterclient" << endl;
}

void
parseargs (int argc, char** argv)
{
	// These are the options we recognize
	static struct option longOptions[] = {
		{ "channels", 1, 0, 'c' },
		{ "remoteip", 1, 0, 'r' },
		{ 0, 0, 0, 0 } // "Terminator"
	};

	channels = 0;
	while (1) {
		int option_index = 0;
		int c = getopt_long_only (argc, argv, \
			"c:r:", longOptions, &option_index);
		if (c == -1) // No more options
			break;
		switch (c) {
			case 'c':
				channels = atoi (optarg);
				bindToPorts = new myIPAddress[channels];
				sendToPorts = new myIPAddress[channels];
				break;
			case 'r':
				peerIP.IPAddress (optarg);
				break;
			default:
				cerr << "unknown option" << endl;
		}
	}

	// All arguments are mandatory
	if (argc != 5 + 2 * channels) {
		cerr << "incorrect number of option options" << endl;
		usage (argv[0]);
		exit (1);
	}

	// Remaining args are the local and remote ports;
	// put them in their respective arrays.
	while (optind < argc) {
		for (int i = 0; i < channels; i++) {
			sendToPorts[i].port (atoi(argv[optind++]));
			sendToPorts[i].IPAddress (peerIP.IPAddress());
		}
		for (int i = 0; i < channels; i++) {
			bindToPorts[i].port (atoi(argv[optind++]));
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
//                                 MAIN
///////////////////////////////////////////////////////////////////////////////

int main (int argc, char** argv) {

	// Remove the leading path from our name
	if (strrchr(argv[0], '/')) {
		argv[0] = strrchr(argv[0], '/');
		argv[0]++;
	}
	parseargs (argc, argv);

	///////////////////////////////////////////////////////////////////////
	// Create the data buffers for each channel with the maximum possible
	// size each; there's obviously some waste here, but that is
	// irrelevant, since it's at most 64K per channel; 100 channels might
	// be accomodated with 6.4Mb of RAM.
	///////////////////////////////////////////////////////////////////////

	// Maximum number of frames that can fit in a network buffer
	jack_nframes_t maxFrames = \
		MAX_NET_BUFSIZE/sizeof (jack_default_audio_sample_t);
	jack_default_audio_sample_t* buffers[channels];
	for (int i = 0; i < channels; i++) {
		buffers[i] = new jack_default_audio_sample_t[maxFrames];
	}

	// Create the network sockets
	udpReadSocket<jack_default_audio_sample_t>* inputSockets[channels];
	udpWriteSocket<jack_default_audio_sample_t>* outputSockets[channels];
	for (int i = 0; i < channels; i++) {
		inputSockets[i] = new udpReadSocket\
			<jack_default_audio_sample_t> (bindToPorts[i], true);
		outputSockets[i] = new udpWriteSocket\
			<jack_default_audio_sample_t> (sendToPorts[i]);
	}

	// Set realtime scheduling
	struct sched_param processingPriority;
	processingPriority.sched_priority = 99;
	sched_setscheduler (0, SCHED_FIFO, &processingPriority);
	// Lock all address space in RAM
	mlockall(MCL_FUTURE);

	// Processing loop
	jack_nframes_t numReadItens = 0;
	for (int i = 0;;i++) {

		// This should be made with poll() or select().
		for (int j = 0; j < channels; j++) {
			numReadItens = inputSockets[j]->read \
				(buffers[j], maxFrames);
			// Just for kicks, lets reduce the amplitude of the
			// signal so we can check that data is really being
			// processed
			// for (int k = 0; j < numReadItens; j++) {
			//	buffers[j][k] *= 4;
			//}
			outputSockets[j]->write (buffers[j], numReadItens);
		}

		// Provide some feedback so the user can tell
		// if we're actually doing something.
		if (i > 100) {
			cout << "mark!" << endl;
			i = 0;
		}
	}
	// Look, ma, no need to dealocate buffers/sockets!
	return 0;
}
