#ifndef MYSOCKET_H
#define MYSOCKET_H

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//    This file defines some *very* simple classes for UDP communication
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// Files that include this file must also include:
// <sys/time.h>, <time.h> - for gettimeofday
// <errno.h> - to check the condition of the non-blocking socket
// <sys/types.h>, <sys/socket.h>, <arpa/inet.h>,
// <fcntl.h>, <netinet/in.h>, <unistd.h> - networking
// <string.h> - memset()
// <strings.h> - C++-style strings are needed by myipaddr.hh
// myipaddr.h


// FIXME - There's no code for dealing with errors here. Also, this is
// definetely NOT the object-oriented way of doing things, so this might
// as well have been coded in C instead of C++.


// Maximum time (in microseconds) that we keep on retrying on a non-blocking
// socket, default is 500ms (which is a *lot*); needed because the process runs
// with priority SCHED_FIFO and could lock the machine in an infinite loop; in
// "The practise of programming", Kernighan & Pike suggest that "enum" is
// better than "#define".
enum { MAX_SOCKET_POLL_TIME = 500000 };


///////////////////////////////////////////////////////////////////////////////
// An idiotic base class for UDP communication; this should be seen
// as a pure virtual class, even if it isn't.
///////////////////////////////////////////////////////////////////////////////
template <class T>
class udpSocket {
public:
	udpSocket (const myIPAddress& IPAddr = "0.0.0.0");
	myIPAddress IPAddress () const;
	virtual ~udpSocket ();
protected:
	int socket;
	struct sockaddr_in address;
};


///////////////////////////////////////////////////////////////////////////////
// A class for sending UDP data to a remote machine
///////////////////////////////////////////////////////////////////////////////
template <class T>
class udpWriteSocket: public udpSocket<T> {
public:
	// Specify where to send data on the constructor
	udpWriteSocket(const myIPAddress& IPAddr);
	inline ssize_t write (const T* buf, size_t count);
};


///////////////////////////////////////////////////////////////////////////////
// A class for receiving UDP data from a remote machine
///////////////////////////////////////////////////////////////////////////////
template <class T>
class udpReadSocket: public udpSocket<T> {
public:
	// Specify address/port to bind to on the constructor.
	udpReadSocket (const myIPAddress& IPAddr = "0.0.0.0", bool blocking = 1);
	inline ssize_t read (T* buf, size_t count);
private:
	bool blocking; // is this a blocking socket?
};


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//        Below this point, implementation; necessary to be on the
//        header file since these are templates.
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Constructors for the three classes; create and "connect" the socket
///////////////////////////////////////////////////////////////////////////////
template <class T>
udpSocket<T>::udpSocket (const myIPAddress& IPAddr = "0.0.0.0") {
	// "initializes" (zeroes) the address struct
	memset (&address, 0, sizeof (sockaddr_in));
	// Now setups the parameters and create the socket; doesn't
	// matter if it's a read or write socket.
	address.sin_family = AF_INET;
	address.sin_port = htons (IPAddr.port());
	if (IPAddr.IPAddress() == "0.0.0.0") {
		address.sin_addr.s_addr = htonl (INADDR_ANY);
	} else {
		inet_pton (AF_INET, IPAddr.IPAddress().c_str(), \
						&address.sin_addr);
	}
	socket = ::socket (PF_INET, SOCK_DGRAM, 0);
};

// Too much semicolons on a single line :-). This just initializes the
// constructor of the superclass (udpSocket, which creates the socket) before
// running the code for this class' constructor. The same happens on the
// constructor for the other class define below.
template <class T>
udpWriteSocket<T>::udpWriteSocket (const myIPAddress& IPAddr) \
			: ::udpSocket<T>::udpSocket (IPAddr) {
	// $%^#@^#%$~%@*[#$!!!! static_cast doesn't work here!
	connect (socket, (struct sockaddr*)(&address), \
		static_cast<socklen_t>(sizeof (address)));
	// After bind/connect, if the address is INADDR_ANY, the "address"
	// variable doesn't hold the correct value for address/port. Let's
	// update it here.
	socklen_t size = sizeof (address);
	getsockname (socket, (struct sockaddr*)(&address), &size);
};

template <class T>
udpReadSocket<T>::udpReadSocket (const myIPAddress& IPAddr = "0.0.0.0", bool blocking) \
			: ::udpSocket<T>::udpSocket (IPAddr) {
	// $%^#@^#%$~%@*[#$!!!! static_cast doesn't work here!
	bind (socket, (struct sockaddr*)(&address), \
		static_cast<socklen_t>(sizeof (address)));
	this->blocking = blocking;
	if (! blocking) {
		// Set non-blocking mode
		fcntl (socket, F_SETFL, fcntl (socket, F_GETFL) | O_NONBLOCK);
	}
	// After bind/connect, if the address is INADDR_ANY, the "address"
	// variable doesn't hold the correct value for address/port. Let's
	// update it here.
	socklen_t size = sizeof (address);
	getsockname (socket, (struct sockaddr*)(&address), &size);
};

///////////////////////////////////////////////////////////////////////////////
// Read and write data
///////////////////////////////////////////////////////////////////////////////
template <class T>
ssize_t
udpWriteSocket<T>::write (const T* buf, size_t count) {
	return ::write (socket, buf, count * sizeof (T)) / sizeof (T);
}

template <class T>
ssize_t
udpReadSocket<T>::read (T* buf, size_t count) {
	// Depending on the way the object was created, this
	// call may block or not.
	ssize_t retval = ::read (socket, buf, count * sizeof (T));

	// If this is a non-blocking call and there is no data to
	// read, busy-wait for "a while" (MAX_SOCKET_POLL_TIME). Other
	// errors are treated "normally". I couldn't find a way to write
	// this loop in a more elegant style...
	if (!blocking) {
		struct timeval startTime, currentTime;
		gettimeofday (&startTime, 0);
		while (retval < 0 && errno == EAGAIN) {
			// FIXME Can this speed up network data processing
			// is it just stupid (I think the latter...)?
			sched_yield ();
			retval  = ::read (socket, buf, count * sizeof (T));
			gettimeofday(&currentTime, 0);
			if ( (currentTime.tv_sec - startTime.tv_sec) * 1000000\
				+ currentTime.tv_usec - startTime.tv_usec > \
				MAX_SOCKET_POLL_TIME) {
				cerr << "socket timeout!!!" << endl;
				return 0;
			}
		}
	} else {
		// The socket "unblocked" and returned some data; now let's
		// change the socket to non-blocking mode temporarily.
		fcntl (socket, F_SETFL, fcntl (socket, F_GETFL) | O_NONBLOCK);
	}
	// Now that we read something, let's make sure we read
	// *everything* that's available, since only the "last"
	// datagram available is useful for us; the others are "late".
	ssize_t testing_retval = retval;
	int droppedPackets = -1;
	while (testing_retval > 0) {
		droppedPackets++;
		retval = testing_retval;
		testing_retval = ::read (socket, buf, count * sizeof (T));
	}
	if (droppedPackets) {
		cerr << "dropped " << droppedPackets << " packets!" << endl;
	}
	if (blocking) {
		// Reset the socket to blocking mode
		fcntl (socket, F_SETFL, fcntl (socket, F_GETFL) ^ O_NONBLOCK);
	}
	return retval / sizeof (T);
}

template <class T>
myIPAddress
udpSocket<T>::IPAddress () const {
	char* str = new char [INET_ADDRSTRLEN]; // maximum size of a IP address
	str[0] = '\0';
	inet_ntop(AF_INET, static_cast<const void*>(&address.sin_addr), \
						str, INET_ADDRSTRLEN);
	myIPAddress IP (str);
	delete str;
	IP.port(ntohs (address.sin_port));
	return IP;
}

///////////////////////////////////////////////////////////////////////////////
// Close the socket on destruction, duh.
///////////////////////////////////////////////////////////////////////////////
template <class T>
udpSocket<T>::~udpSocket () {
	close (socket);
};

#endif // ifndef MYSOCKET_H
